#  Copyright 2019 Michael Kemna.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from enum import Enum

from sqlalchemy import text


class Query(Enum):
    """
    This Enum is used to create a framework to store the queries that are used by the application.
    Of course, you can also create your own Query Enum in your own module.

    source to execute raw sql with sqlalchemy:
    https://chartio.com/resources/tutorials/how-to-execute-raw-sql-in-sqlalchemy/
    """
    select_forecast_24h = text("""
        SELECT r.id, lat, lng
        FROM resort as r
        LEFT JOIN (SELECT id, resort_id
                   FROM forecast
                   WHERE date = current_date
                   and timepoint = 0
                   ) as f on r.id = f.resort_id
        WHERE f.id is NULL
    """)

    select_weather_3h = text("""
        SELECT r.id, lat, lng
        FROM resort as r
        LEFT JOIN (SELECT id, resort_id 
                   FROM weather 
                   WHERE date_request > current_timestamp - 3 * interval '1 hour'
                   ) as w on r.id = w.resort_id
        WHERE w.id is NULL
    """)



#  Copyright 2019 Michael Kemna.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from contextlib import contextmanager
from typing import List

from sqlalchemy import create_engine, MetaData, Table, Column
from sqlalchemy.engine import Engine, Connection, ResultProxy

from powderbooking.models import model_resort, model_weather, model_forecast, model_forecast_week
from powderbooking.query import Query


class DatabaseHandler:
    """
    Database handler to manage all powderbooking interactions in one place.

    inspired by: https://github.com/rshk/flask-sqlalchemy-core/
    """
    engine: Engine
    metadata: MetaData

    def __init__(self, database_url: str):
        self.engine, self.metadata = self.build_postgres_database(database_url=database_url)

    @staticmethod
    def build_postgres_database(database_url: str) -> (Engine, MetaData):
        engine = create_engine(database_url)
        metadata = MetaData()

        model_resort(metadata)
        model_weather(metadata)
        model_forecast(metadata)
        model_forecast_week(metadata)

        metadata.create_all(engine)

        return engine, metadata

    @contextmanager
    def connect(self) -> Connection:
        """
        Connect to the powderbooking, ensures that the connection is closed once all transactions have occurred.
        """
        with self.engine.connect() as conn:
            yield conn
            conn.close()

    def execute(self, *args, **kwargs) -> ResultProxy:
        """
        Execute the inserted args and kwargs onto the powderbooking and return the ResultProxy.
        """
        with self.connect() as conn:
            return conn.execute(*args, **kwargs)

    def execute_query(self, query: Query, *args, **kwargs) -> ResultProxy:
        """
        Execute the pre-specified query.

        :param query: the Query that should be executed.
        :param args: any other positional arguments.
        :param kwargs: any other keyword arguments.
        :return: the ResultProxy.
        """
        return self.execute(query.value, *args, **kwargs)

    def get_table(self, table_name: str) -> Table:
        """
        Get the table from the metadata.

        :param table_name: the name of the table that should be retrieved
        :return: the Table object.
        """
        if table_name not in self.metadata.tables.keys():
            raise ValueError(f'{table_name} not in the tables of powderbooking')
        return self.metadata.tables[table_name]

    def get_table_column(self, table_name: str, column_name: str) -> Column:
        """
        Get the column of the particular table from the metadata.

        :param table_name: the name of the table that
        :param table_name: the name of the column that should be retrieved
        :return: the Column object.
        """
        table = self.get_table(table_name=table_name)
        if column_name not in table.columns:
            raise ValueError(f'{column_name} not in the columns of the table {table_name} of powderbooking')
        return table.columns[column_name]

    def insert(self, table_name: str, values: List[dict]) -> ResultProxy:
        """
        Insert the inserted list of values into the table that is given.
        Will raise a ValueError if the table is not inside the powderbooking.

        :param table_name: the name of the table that should be inserted into.
        :param values: a list of new rows that should be inserted.
        :return: the ResultProxy.
        """
        return self.execute(self.get_table(table_name).insert(), values)



#  Copyright 2019 Michael Kemna.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from sqlalchemy import Table, Column, Integer, Float, Date, DateTime, String, Sequence, MetaData, \
    ForeignKey, UniqueConstraint


def model_resort(metadata: MetaData) -> Table:
    return Table('resort', metadata,
                 Column('id', Integer, Sequence('resort_id_seq'), primary_key=True,
                        comment='The identifier of the resort'),
                 Column('continent', String, comment='The continent where the resort is located'),
                 Column('name', String, comment='The shorthand name of the resort'),
                 Column('fullname', String, comment='The full name of the resort'),
                 Column('country', String, comment='The country where the resort is located'),
                 Column('village', String, comment='The village where the resort is located'),
                 Column('lat', Float, comment='The latitudinal coordinate of the geolocation of the resort'),
                 Column('lng', Float, comment='The longitudinal coordinate of the geolocation of the resort'),
                 Column('altitude_min_m', Integer, comment='The lowest altitude of the resort (in metres)'),
                 Column('altitude_max_m', Integer, comment='The highest altitude of the resort (in metres)'),
                 Column('lifts', Integer, comment='The amount of lifts in the resort'),
                 Column('slopes_total_km', Integer, comment='The length of all slopes in the resort (in kilometres)'),
                 Column('slopes_blue_km', Integer, comment='The length of blue slopes in the resort (in kilometres)'),
                 Column('slopes_red_km', Integer, comment='The length of red slopes in the resort (in kilometres)'),
                 Column('slopes_black_km', Integer, comment='The length of black slopes in the resort (in kilometres)'),
                 UniqueConstraint('village', 'country'),
                 )


def model_weather(metadata: MetaData) -> Table:
    return Table('weather', metadata,
                 Column('id', Integer, Sequence('weather_id_seq'), primary_key=True,
                        comment='The identifier of the weather report'),
                 Column('resort_id', Integer, ForeignKey('resort.id', onupdate='CASCADE', ondelete='CASCADE'),
                        comment='Foreign key to the related resort'),
                 Column('date_request', DateTime, comment='The date on which the weather was requested by the scraper'),
                 Column('dt', DateTime, comment='The datetime as given by the scraped api'),
                 Column('date', Date, comment='The date derived from dt'),
                 Column('timepoint', Integer, default=-1,
                        comment='Ordinal timepoint of the day, categorized every 3 hours (range 0-7)'),
                 Column('temperature_c', Float, comment='The temperature in Celsius'),
                 Column('wind_speed_kmh', Float, comment='The wind speed in kilometres per hour'),
                 Column('wind_direction_deg', Float, comment='The wind direction in degrees'),
                 Column('visibility_km', Float, comment='The visibility in kilometres'),
                 Column('clouds_pct', Float, comment='The amount of clouds in percentage'),
                 Column('snow_3h_mm', Float, comment='The amount of snow in the last 3 hours in millimetres'),
                 Column('rain_3h_mm', Float, comment='The amount of rain in the last 3 hours in millimetres'),
                 UniqueConstraint('date', 'timepoint', 'resort_id'),
                 )


def model_forecast(metadata: MetaData) -> Table:
    return Table('forecast', metadata,
                 Column('id', Integer, Sequence('forecast_id_seq'), primary_key=True),
                 Column('resort_id', Integer, ForeignKey('resort.id', onupdate="CASCADE", ondelete="CASCADE")),
                 Column('date_request', DateTime),
                 Column('date', Date),
                 Column('timepoint', Integer),
                 Column('temperature_max_c', Float),
                 Column('temperature_min_c', Float),
                 Column('rain_total_mm', Float),
                 Column('snow_total_mm', Float),
                 Column('prob_precip_pct', Float),
                 Column('wind_speed_max_kmh', Float),
                 Column('windgst_max_kmh', Float),
                 UniqueConstraint('date', 'timepoint', 'resort_id'),
                 )


def model_forecast_week(metadata: MetaData) -> Table:
    return Table('forecast_week', metadata,
                 Column('id', Integer, Sequence('forecast_week__id_seq'), primary_key=True),
                 Column('resort_id', Integer, ForeignKey('resort.id', onupdate="CASCADE", ondelete="CASCADE"),
                        unique=True),
                 Column('date_request', DateTime),
                 Column('date', Date),
                 Column('rain_week_mm', Float),
                 Column('snow_week_mm', Float),
                 )



from setuptools import setup


setup(name='powderbooking',
      version='0.4.1',
      description='Application to show the best hotels with the weather',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3.6',
          'Topic :: Database :: Database Engines/Servers',
      ],
      keywords='powderbooking powder snow booking hotel powderbooking model',
      url='http://github.com/mrasap/powderbooking-database',
      author='mrasap',
      author_email='michael.kemna@gmail.com',
      license='Apache',
      packages=['powderbooking'],
      install_requires=[
          'SQLAlchemy',
          'psycopg2-binary',
      ],
      include_package_data=True,
      zip_safe=False)
